from . import config

from datetime import date, time, datetime


def parsedatetime(date, time):
    """# DD.MM.YY, HH:MM"""
    return datetime.strptime("{}-{}".format(date, time), "%d.%m.%y-%H:%M")


class Coordinate:
    """Journey Planner coordinate.
    Converts between internal xy coords and regular lat/lon
    """

    def __init__(self, *, latitude, longitude):
        self.latitude = latitude
        self.longitude = longitude

    @property
    def x(self):
        return self._z(self.longitude)

    @property
    def y(self):
        return self._z(self.latitude)

    @staticmethod
    def _z(foo):
        return int(foo * config.COORDINATE_MULTIPLIER)

    @classmethod
    def fromxy(cls, *, x, y):
        return cls(
            latitude=y / config.COORDINATE_MULTIPLIER,
            longitude=x / config.COORDINATE_MULTIPLIER,
        )


class Endpoint(Coordinate):
    def __init__(self, *, latitude, longitude, name=None):
        super().__init__(latitude=latitude, longitude=longitude)
        self.name = name


class Resource:
    def __repr__(self):
        return repr(self.__dict__)


class Location(Resource):
    def __init__(self, element):
        self.name = element.get("name")
        coordinate = Coordinate.fromxy(y=int(element.get("y")), x=int(element.get("x")))
        self.latitude = coordinate.latitude
        self.longitude = coordinate.longitude

        if element.tag == "StopLocation":
            self.type = "stop"
            self.id = int(element.get("id"))

        if element.tag == "CoordLocation":
            self.id = None
            if element.get("type") == "ADR":
                self.type = "address"
            if element.get("type") == "POI":
                self.type = "poi"


class Stop(Resource):
    def __init__(self, element):
        self.id = int(element.get("id"))
        self.name = element.get("name")
        c = Coordinate.fromxy(y=int(element.get("y")), x=int(element.get("x")))
        (self.latitude, self.longitude) = (c.latitude, c.longitude)

        try:
            self.distance = int(element.get("distance"))
        except ValueError:
            self.distance = None


class LegEndpoint(Resource):
    def __init__(self, element):
        self.datetime = parsedatetime(element.get("date"), element.get("time"))

        self.name = element.get("name")
        self.type = element.get("type")
        self.track = element.get("track")
        self.realtrack = element.get("rtTrack")

    @property
    def delayed(self):
        return self.realtime.datetime


class Leg(Resource):
    def __init__(self, leg):
        self.name = leg.get("name")
        self.type = leg.get("type")

        self.origin = LegEndpoint(leg.find("Origin"))
        self.destination = LegEndpoint(leg.find("Destination"))

        try:
            self.notes = leg.find("Notes").get("text")
        except AttributeError:
            self.notes = None

        try:
            self.ref = leg.find("JourneyDetailRef").get("ref")
        except AttributeError:
            self.ref = None

    @property
    def duration(self):
        """Leg travel duration.
        """
        return self.destination.datetime - self.origin.datetime


class Trip(Resource):
    def __init__(self, element):
        self.alternative = element.get("alternative", False)
        self.valid = element.get("valid", True)
        self.legs = [Leg(leg) for leg in element]

    @property
    def duration(self):
        """Duration of entire trip.
        """
        durations = iter(leg.duration for leg in self.legs)
        first = next(durations)
        return sum(durations, first)



from .request import Requester
from .type import Location, Stop, Trip, Coordinate, Endpoint


class JourneyPlanner:
    def __init__(self):
        self.requester = Requester()

    def authenticate(self, username, password):
        self.requester.authenticate(username, password)

    def location(self, query):
        """Fuzzy search for stops/stations/points of interest/adresses.

        The location service can be used to perform a pattern matching of a
        user input and to retrieve a list of possible matches in the journey
        planner database. Possible matches might be stops/stations, points of
        interest and addresses.
        """
        tree = self.requester.get("locations", input=query)

        yield from (Location(element) for element in tree)

    def stopsnearby(self, *, latitude, longitude, radius=None, limit=None):
        """Find stops within a radius of a given coordinate.

        Each stop location contains the name of the
        station/stop, the coordinate, the id and the distance from the request
        coordinate in meters. All distances are as the crow flies and not
        routed.
        """
        coordinate = Coordinate(latitude=latitude, longitude=longitude)
        tree = self.requester.get(
            "stopsNearby",
            coordX=coordinate.x,
            coordY=coordinate.y,
            maxRadius=radius,
            maxNumber=limit,
        )

        yield from (Stop(element) for element in tree)

    def trip(
        self,
        origin,
        destination,
        *,
        via=None,
        date=None,
        time=None,
        arrivals=False,
        bus=True,
        metro=True,
        train=True
    ):
        """The trip service calculates a trip from a specified origin to a
        specified destination. These might be stop/station IDs or coordinates
        based on addresses and points of interest validated by the location
        service or coordinates freely defined by the client.

        The parameters are named either originId or originCoordX, originCoordY,
        and originCoordName. For the destination the parameters are named either
        destId or destCoordX, destCoordY and destCoordName
        """
        endpointparams = dict()
        endpoints = dict(origin=origin, dest=destination)
        for key, endpoint in endpoints.items():
            try:
                (x, y, name) = (endpoint.x, endpoint.y, endpoint.name)
                endpointparams.update(
                    {
                        "{}CoordX".format(key): x,
                        "{}CoordY".format(key): y,
                        "{}CoordName".format(key): name if name else key,
                    }
                )
            except AttributeError:
                endpointparams["{}Id".format(key)] = endpoint

        if date is not None:
            date = date.strftime("%d.%m.%y")

        tree = self.requester.get(
            "trip",
            date=date,
            time=time,
            viaId=via,
            searchForArrival=int(arrivals),
            useBus=int(bus),
            useMetro=int(metro),
            useTrain=int(train),
            **endpointparams
        )

        yield from (Trip(element) for element in tree)

    def _stationboard(
        self, service, id, *, date=None, time=None, bus=True, metro=True, train=True
    ):
        """The station board board can be retrieved by the service
        departureBoard. This method will return the next 20 departures (or less
        if not existing) from a given point in time.

        In addition to departure boards the service arrivalBoard delivers
        arriving journeys at a specified stop. The parameters are identical to
        the parameters of the departureBoard service.
        """
        tree = self.requester.get(
            service,
            id=id,
            date=date,
            time=time,
            useBus=int(bus),
            useMetro=int(metro),
            useTrain=int(train),
        )

        yield from tree

    def arrivalboard(self, *args, **kwargs):
        return self._stationboard("arrivalBoard", *args, **kwargs)

    def departureboard(self, *args, **kwargs):
        return self._stationboard("departureBoard", *args, **kwargs)

    def multidepartureboard(
        self, *ids, date=None, time=None, bus=True, metro=True, train=True
    ):
        """The multi departure board is a combined departure board for up to 10
        different stops. It can be retrieved by a service called
        multiDepartureBoard. This method will return the next 20 departures (or
        less if not existing) of the defined stops from a given point in time.
        """
        if len(ids) > 10:
            raise Exception("MultiDepartureBoard does not support more than 10 ids")

        tree = self.requester.get(
            "multiDepartureBoard",
            date=date,
            time=time,
            useBus=int(bus),
            useMetro=int(metro),
            useTrain=int(train),
        )

        yield from tree

    def journeydetail(self, reference):
        """The journeyDetail service will deliver information about the complete
        route of a vehicle. This service can’t be called directly but only by
        reference URLs in a result of a trip or departureBoard request. It
        contains a list of all stops/stations of this journey including all
        departure and arrival times (with realtime data if available) and
        additional information like specific attributes about facilities and
        other texts.
        """
        tree = self.requester.get("journeyDetail", ref=reference)

        yield from tree



def fromstring(message):
    """Convert a Rejseplanen API error message into an appropriate exception.
  """
    error_code = message.split()[0]
    message = ERRORS.get(error_code, message)
    first_char = error_code[0]

    if first_char == "R":
        return RestError(message)
    elif first_char == "S":
        return ServerError(message)
    elif first_char == "H":
        return TripServiceError(message)
    else:
        raise JourneyPlannerError(message)


# -------------------------------------------------------------------------------

ERRORS = {
    # General ReST Request Errors
    "R0001": "Unknown service",
    "R0002": "Invalid or missing request parameters",
    "R0007": "Internal communication error",
    # Backend Server Errors
    "S1": "The desired connection to the server could not be established or was not stable.",
    # Connection Request Errors
    "H9380": "Dep./Arr./Intermed defined more than once",
    "H9360": "Error in date field",
    "H9320": "The input is incorrect or incomplete",
    "H9300": "Unknown arrival station",
    "H9280": "Unknown intermediate station",
    "H9260": "Unknown departure station",
    "H9240": "Unsuccessful search",
    "H9230": "An internal error occurred",
    "H9220": "Nearby to the given address stations could not be found",
    "H900": "Unsuccessful or incomplete search (timetable change)",
    "H892": "Inquiry too complex (try entering less intermediate stations)",
    "H891": "No route found (try entering an intermediate station)",
    "H890": "No connections found",
    "H895": "Departure/Arrival are too near",
}


class JourneyPlannerError(Exception):
    """There was an exception that occurred while handling your request."""

    pass


class AuthenticationError(JourneyPlannerError):
    """Authentication failed."""

    pass


class RestError(JourneyPlannerError):
    """General rest error occured"""

    pass


class ServerError(JourneyPlannerError):
    """Backend Server gave error"""

    pass


class TripServiceError(JourneyPlannerError):
    """There was an error with your trip request"""
    pass




from . import JourneyPlanner

import argparse

parser = argparse.ArgumentParser(
    prog="python -m journeyplanner",
    description="Make requests to the journeyplanner API",
)

parser.add_argument("-u", help="authenticate request")
parser.add_argument("-p", help="authenticate request")

subparsers = parser.add_subparsers(help="api service", dest="service")

parser_location = subparsers.add_parser("location", help="location service")
parser_location.add_argument("query", type=str, help="the search query")

parser_trip = subparsers.add_parser("trip", help="trip service")
exorigin = parser_trip.add_mutually_exclusive_group(required=True)
exorigin.add_argument("--originid", type=int, help="origin")
exorigin.add_argument(
    "--origin", nargs=2, type=float, metavar=("LATITUDE", "LONGITUDE")
)
exdest = parser_trip.add_mutually_exclusive_group(required=True)
exdest.add_argument("--destinationid", type=int, help="destination")
exdest.add_argument(
    "--destination", nargs=2, type=float, metavar=("LATITUDE", "LONGITUDE")
)

parser_stopsnearby = subparsers.add_parser("stopsnearby", help="stops nearby service")
parser_stopsnearby.add_argument("latitude", type=float, help="latitude")
parser_stopsnearby.add_argument("longitude", type=float, help="longitude")

jp = JourneyPlanner()
args = parser.parse_args()
if args.u and args.p:
    jp.authenticate(args.u, args.p)

if args.service == "location":
    for x in jp.location(args.query):
        print(str(x))

if args.service == "stopsnearby":
    for x in jp.stopsnearby(
        latitude=args.latitude, longitude=args.longitude, radius=10000
    ):
        print(str(x))

if args.service == "trip":
    origin = args.originid if args.originid else args.origin
    destination = args.destinationid if args.destinationid else args.destination

    for x in jp.trip(origin=origin, destination=destination):
        print(str(x))



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="journeyplanner",
    version="1.0.1",
    license="MIT",
    author="Bytes & Brains",
    author_email="developers@bytesandbrains.com",
    description="Journey Planner API client library",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/bytesandbrains/journeyplanner-python",
    packages=["journeyplanner"],
    test_suite="test",
)
